/*
 * (C) 2005  Alexander Shaposhnikov  <shaposh@isp.nsc.ru>
 *
 *  Licensed under the terms of the GNU GPL License version 2.
 *  Based upon reverse engineered information
 *
 *  BIG FAT DISCLAIMER: Work in progress code. Possibly *dangerous*
 */


#include "cpufreq_nForce4.h"

/*Set Control Register of Specific CPU
*/
int SET_CTL_MSR(struct K8_CTL_MSR ctl_msr)
{
	int err;	
	u32 eax,edx;
	err=do_rdmsr(ctl_msr.cpu,MSR_FIDVID_CTL,&eax,&edx);
	if(err)
	{	
		printk(KERN_ERR "cpufreq: error reading msr %X for CPU %d\n",
				MSR_FIDVID_CTL,ctl_msr.cpu);
		return err;
	}

	SET_C_LO_NEW_FID(&eax,ctl_msr.fid);
	SET_C_LO_NEW_VID(&eax,ctl_msr.vid);
	SET_C_HI_STP_GNT_TO(&edx,ctl_msr.stp_gnt_to);
	if(ctl_msr.init_bit)
		SET_C_LO_INIT(&eax);

	err=do_wrmsr(ctl_msr.cpu,MSR_FIDVID_CTL,eax,edx); 

	if(err)
	{	
		printk(KERN_ERR "cpufreq: error writing msr %X edx %X eax %X for CPU %d\n",
				MSR_FIDVID_CTL,edx,eax,ctl_msr.cpu);
   		return err;
	}
	return 0;
}

/*Get current STATUS

*/
int GET_STATUS_MSR(struct K8_STATUS_MSR *status_msr)
{
	int err;	
	u32 eax,edx;
	err=do_rdmsr(status_msr->cpu,MSR_FIDVID_STATUS,&eax,&edx);
	if(err)
	{	
		printk(KERN_ERR "cpufreq: error reading msr %X for CPU %d\n", 
				MSR_FIDVID_STATUS,status_msr->cpu);
		return err;
	}
	status_msr->current_fid    = GET_S_LO_CURRENT_FID(eax);
	status_msr->start_fid      = GET_S_LO_START_FID(eax);
	status_msr->max_fid        = GET_S_LO_MAX_FID(eax);
	status_msr->max_ramp_vid   = GET_S_LO_MAX_RAMP_VID(eax);
	status_msr->change_pending = GET_S_LO_CHANGE_PENDING(eax);

	status_msr->current_vid    = GET_S_HI_CURRENT_VID(edx);	
	status_msr->start_vid      = GET_S_HI_START_VID(edx);
	status_msr->max_vid        = GET_S_HI_MAX_VID(edx);
	status_msr->min_vid        = GET_S_HI_MIN_VID(edx);
	if(status_msr->min_vid<status_msr->max_vid)
		status_msr->min_vid=K8_MIN_VID;
	return 0;  		   				

}

/*Return real possible final fsb from target, cos there is some 
	granularity
*/
int GET_REAL_TARGET_FSB(u32 target_fsb, u32 *real_target_fsb)
{
	struct PLL in_pll;
	u32 PLL_DIV;
	u32 target_mul;

	if(ALG==0)		  //nTune
	{
		*real_target_fsb=FIND_BEST_PLL(target_fsb,&in_pll);	
		return 0;
	}
	 //clockgen
	
	PLL_DIV=0x100;
	if(ALG==2)
	    PLL_DIV=0x200;
	else    
		if(ALG==3)    
		    PLL_DIV=0x300;
	
	READ_PLL(&in_pll);

	if(in_pll.div!=PLL_DIV) //set div to PLL_DIV 
	{
		if(PLL_DIV*in_pll.mul/in_pll.div>0x3FF)
		{
			printk(KERN_ERR "cpufreq: change ALG to lower value! ALG=%d\n", ALG);
			return -1;
		}
	}
	target_mul=target_fsb*PLL_DIV/PLL_REF_CLOCK;
			
	if(target_mul>0x3FF)
	{
		printk(KERN_ERR "cpufreq: change ALG to lower value! ALG=%d\n", ALG);
		return -1;
	}
	*real_target_fsb=target_mul*PLL_REF_CLOCK/PLL_DIV;
	
        return 0;
}

/*Returns current FSB is KHz.Until i could find how to read current FSB regardless 
	of PLL value, this function can return wrong value if PLL_ENABLE is not set
*/
static int GET_FSB(u32 *fsb)
{	
	struct PLL in_pll;    
	READ_PLL(&in_pll);
	*fsb=in_pll.mul*PLL_REF_CLOCK/in_pll.div;
	return 0;
} 

/*Set FSB.
 ALG=0: Like nTune. 
 ALG=1: Like clockgen. PLL divider is set to 0x100 and only PLL multiplier varied.
 ALG=2: PLL divider is set to 0x200. 
 ALG=3: PLL divider is set to 0x300. 
*/
int SET_FSB(u32 target_fsb)
{
	struct PLL pll;	

	
	if(ALG>3)
	    return -1;
	READ_PLL(&pll);
	
	if(ALG==0)	 //nTune   
	{	
		u32 best_fsb;
		u32 current_fsb;
    
		current_fsb=PLL_REF_CLOCK*pll.mul/pll.div;						
		best_fsb=FIND_BEST_PLL(target_fsb,&pll);
				
		//We change fsb by steps of PLL_FSB_STEP at a time
		while( ABS(current_fsb-best_fsb)>PLL_FSB_STEP)
		{
			if(current_fsb<best_fsb)
				current_fsb+=PLL_FSB_STEP;
			else
				current_fsb-=PLL_FSB_STEP;
						
			FIND_BEST_PLL(current_fsb,&pll);
			WRITE_PLL(pll);
		}
		FIND_BEST_PLL(target_fsb,&pll);
		WRITE_PLL(pll);
	}
	else  //clockgen	
	{	
		u32 PLL_DIV;
		u32 target_mul;
				
		PLL_DIV=0x100;
	
		if(ALG==2)
			PLL_DIV=0x200;
		else
		    if(ALG==3)	    	
			PLL_DIV=0x300;
	
		if(pll.div!=PLL_DIV) //set pll.div to PLL_DIV 
		{
			pll.mul=PLL_DIV*pll.mul/pll.div;
			if(pll.mul>0x3FF)
			{
				printk(KERN_ERR "cpufreq: change ALG to lower value! ALG=%d\n", ALG);
				return -1;
			}
			pll.div=PLL_DIV;
			SET_PLL(&pll);
			WRITE_PLL(pll);
		}
		target_mul=target_fsb*PLL_DIV/PLL_REF_CLOCK;
		if(target_mul>0x3FF)
		{
			printk(KERN_ERR "cpufreq: change ALG to lower value! ALG=%d\n", ALG);
			return -1;
		}
		while(pll.mul!=target_mul)
		{
			if(target_mul>pll.mul)
				pll.mul++;	
			else
				pll.mul--;
			SET_PLL(&pll);
			WRITE_PLL(pll);
		}
	}
	return 0;
} 

static int CK804_target(struct cpufreq_policy *policy,
			  unsigned int target_freq, unsigned int relation)
{
	u32 real_target_fsb;
	u32 target_fsb;
	u32 current_fsb, old_fsb;
	int cpu,ret;
	struct cpufreq_freqs freqs;
	struct K8_STATUS_MSR status_msr;

	if(!cpu_online(policy->cpu))
		return -ENODEV;

	
	if ((target_freq > policy->max) || (target_freq < policy->min))
		return -EINVAL;

	ret=0;
	status_msr.cpu=policy->cpu;
	GET_STATUS_MSR(&status_msr);
	
	target_fsb=target_freq/(FID_to_MUL(status_msr.current_fid));    
	GET_REAL_TARGET_FSB(target_fsb, &real_target_fsb);
	
	GET_FSB(&current_fsb);

	if (real_target_fsb==current_fsb)
		return 0;

	old_fsb=current_fsb;

   /*We have to send notifiers with all CPUs!*/
	for_each_cpu_mask(cpu,AFFECTED_CPUS)
	{
		u32 mul;
		status_msr.cpu=cpu;
		GET_STATUS_MSR(&status_msr);
		mul=FID_to_MUL(status_msr.current_fid);
		freqs.old = old_fsb*mul;
		freqs.new = real_target_fsb*mul;
		freqs.cpu = cpu;
		cpufreq_notify_transition(&freqs,CPUFREQ_PRECHANGE);
	}		

	// Disable IRQs ?

	preempt_disable();
	if (SET_FSB(target_fsb) < 0)
	{
		printk(KERN_ERR "cpufreq: FSB changing from %d to %d failed\n",old_fsb,real_target_fsb);	
		ret=-1;
	}
	else
	{
		printk(KERN_INFO "cpufreq: FSB changed from %d to %d\n",old_fsb, real_target_fsb);	
	    		
	}
	preempt_enable();    	       

	GET_FSB(&current_fsb);

	for_each_cpu_mask(cpu,AFFECTED_CPUS)
	{

		u32 mul;
		status_msr.cpu=cpu;
		GET_STATUS_MSR(&status_msr);
		mul=FID_to_MUL(status_msr.current_fid);
		freqs.old = old_fsb*mul;
		freqs.new = current_fsb*mul;
		freqs.cpu = cpu;
		cpufreq_notify_transition(&freqs, CPUFREQ_POSTCHANGE);
        }
	return ret;
}
static int CK804_verify(struct cpufreq_policy *policy)
{
	struct K8_STATUS_MSR status_msr;
	u32 min,max;    

	if(!cpu_online(policy->cpu))
		return -ENODEV;


	status_msr.cpu=policy->cpu;
	GET_STATUS_MSR(&status_msr);


	min = PLL_MIN_FSB*FID_to_MUL(status_msr.current_fid);
	if(ALG!=3) //PLL_DIV!=0x300
	{
		max = PLL_MAX_FSB*FID_to_MUL(status_msr.current_fid);
	}
	else
	{
		max=0x3FF*PLL_REF_CLOCK*FID_to_MUL(status_msr.current_fid)/0x300;
	}
	
	
	cpufreq_verify_within_limits(policy,min,max);

	return 0;
}
static unsigned int CK804_get(unsigned int cpu)
{
	u32 current_fsb;
	u32 current_cpu_clock;    
	struct K8_STATUS_MSR status_msr;
	
	if(!cpu_online(cpu))
		return -ENODEV;


	GET_FSB(&current_fsb);

	status_msr.cpu=cpu;
	GET_STATUS_MSR(&status_msr);
	current_cpu_clock=current_fsb*FID_to_MUL(status_msr.current_fid);
	
	return current_cpu_clock ;
}
static int CK804_cpu_init(struct cpufreq_policy *policy)
{
	struct K8_STATUS_MSR status_msr;


	if(!cpu_online(policy->cpu))
		return -ENODEV;

	status_msr.cpu=policy->cpu;
	GET_STATUS_MSR(&status_msr);

	policy->cpus=AFFECTED_CPUS;
	
	policy->cpuinfo.min_freq = PLL_MIN_FSB*FID_to_MUL(K8_MIN_FID);
	if(ALG!=3)
	{
		policy->cpuinfo.max_freq = PLL_MAX_FSB*FID_to_MUL(status_msr.max_fid);
	}
	else
	{
		policy->cpuinfo.max_freq = 0x3FF*PLL_REF_CLOCK*FID_to_MUL(status_msr.max_fid)/0x300;
	}

	policy->cpuinfo.transition_latency = CPUFREQ_ETERNAL;
	policy->cur = START_FSB*FID_to_MUL(status_msr.current_fid);
	if( (START_FSB-PLL_SAFE_DISTANCE)>PLL_MIN_FSB )
		policy->min = (START_FSB-PLL_SAFE_DISTANCE)*FID_to_MUL(status_msr.current_fid);	
	else
		policy->min = PLL_MIN_FSB*FID_to_MUL(status_msr.current_fid);	
	
    		
	policy->max = START_FSB*FID_to_MUL(status_msr.current_fid);

	policy->governor = CPUFREQ_DEFAULT_GOVERNOR;	
	return 0;
}
static int CK804_cpu_exit(struct cpufreq_policy *policy)
{
	return 0;
}

static struct cpufreq_driver CK804_driver = {
	.name   = "cpufreq_nForce4",
	.verify = CK804_verify,
	.target = CK804_target,
	.get    = CK804_get,
	.init   = CK804_cpu_init,
	.exit   = CK804_cpu_exit,
	.owner  = THIS_MODULE,
};


/**
  nForce4_detect_isa_bridge - detect the isa bridge 
to get PLL port 
 */
static unsigned int CK804_detect_isa_bridge(void)
{
	u8 revision;

	CK804_ISA_BRIDGE_DEV = pci_get_subsys(PCI_VENDOR_ID_NVIDIA,
                                             PCI_DEVICE_ID_NVIDIA_CK804_PRO,
                                             PCI_ANY_ID,
                                             PCI_ANY_ID,
                                             NULL);

	if (CK804_ISA_BRIDGE_DEV == NULL)
		return -ENODEV;

	pci_read_config_byte(CK804_ISA_BRIDGE_DEV, PCI_REVISION_ID, &revision);

	printk(KERN_INFO "cpufreq: Detected CK804 ISA bridge  revision %X\n",
	       revision);
	printk(KERN_INFO "cpufreq: FSB changing is dangerous and can lead to crashes and data loss\n");

	return 0;
}


/**
 * CK804_init - initializes the CK804 CPUFreq driver
 *
 */
static int __init CK804_init(void)
{	
	int cpu;
	u16 tmp;
	
	/* detect chipset */
	if (CK804_detect_isa_bridge()) {
		printk(KERN_ERR "cpufreq: No CK804 ISA bridge detected.\n");
		return -ENODEV;
	}
	pci_read_config_word(CK804_ISA_BRIDGE_DEV, PLL_ADR_REG, &tmp);
	PLL_PORT=(tmp&0xff00)+0x08;
	PLL_ENABLE_PORT=(tmp&0xff00)+0xC8;
	ENABLE_PLL();
	GET_FSB(&START_FSB);
	
	for(cpu=0;cpu<NR_CPUS;cpu++)
	    if(cpu_online(cpu))
		cpu_set(cpu,AFFECTED_CPUS);
	
	//TODO: fix this, have to send notifiers on FID change ?
	for_each_cpu_mask(cpu,AFFECTED_CPUS)
	{
		struct K8_STATUS_MSR status_msr;
		struct K8_CTL_MSR ctl_msr;

		status_msr.cpu=cpu;			
		GET_STATUS_MSR(&status_msr);		
	
		if((START_VID!=UNSPECIFIED)&&(status_msr.current_vid!=START_VID))
		{
			/*Need to set VID first*/		
			ctl_msr.cpu=cpu;
			ctl_msr.fid=status_msr.current_fid;
			ctl_msr.vid=START_VID;
			ctl_msr.stp_gnt_to=DEF_STP_GNT_TO;
			ctl_msr.init_bit=1;
			SET_CTL_MSR(ctl_msr);			
		    	GET_STATUS_MSR(&status_msr);	
		}	
		if((START_MUL!=UNSPECIFIED)&&(status_msr.current_fid!=MUL_to_FID(START_MUL)))
		{
			/*Need to set FID */
			ctl_msr.cpu=cpu;
			ctl_msr.vid=status_msr.current_vid;
			ctl_msr.fid=MUL_to_FID(START_MUL);		
			ctl_msr.stp_gnt_to=DEF_STP_GNT_TO;
			ctl_msr.init_bit=1;	    		
			SET_CTL_MSR(ctl_msr);
		    	GET_STATUS_MSR(&status_msr);				
    		}
		PRINT_STATUS_MSR(status_msr);
	}
	return cpufreq_register_driver(&CK804_driver);
}

/**
 * CK804_exit - unregisters cpufreq module
 *
 *   Unregisters CK804 FSB change support.
 */
static void __exit CK804_exit(void)
{	
	printk(KERN_INFO "cpufreq: Exiting\n");	
	cpufreq_unregister_driver(&CK804_driver); 
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Alexander Shaposhnikov <shaposh@isp.nsc.ru>");
MODULE_DESCRIPTION("nVidia CK804 (nForce4) FSB changing cpufreq driver ver 0.5");

//START_VID:Set voltage in millivolts=1550-25*START_VID
module_param(START_VID,int,0444);
//START_MUL: Set multiplier
module_param(START_MUL,int,0444);
//ALG: ALG=0 default. Seems to be the best. In case of instabilities, try 1,2,3. 
//2,3 is not tested. 0=nTune, precise FSB control. 1=Clockgen.
module_param(ALG,int,0444);

module_init(CK804_init);
module_exit(CK804_exit);
