//0.1 -Intitial release
//0.2 -Some bug fixing and code clearing 
//0.3 -Major bug fix! PLL decoding was totally wrong before.
// Surprisingly, it worked somehow:)
//0.4 -Add nTune -like overclocking, with precise FSB value
//0.5 -Minor bug fixes
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/cpufreq.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/smp.h>
#include <asm/msr.h>

#define DRV_VERSION 	                "0.5"			
#define UNSPECIFIED                     0xFFFFFFFF
#define PLL_MIN_FSB	                100000
#define PLL_MAX_FSB	                400000		
#define PLL_DEFAULT_FSB                 200000
#define PLL_SAFE_DISTANCE               50000
#define PLL_FSB_STEP			1000
#define PLL_MIN_MUL                     10
#define PLL_MIN_DIV                     10
#define PCI_DEVICE_ID_NVIDIA_CK804_PRO  0x0051

/*CK804 PLL related items 
  TODO: decode entire PLL reg
*/

#define PLL_ADR_REG         0x68	
#define PLL_REF_CLOCK 	    193752
//#define PLL_DELAY           5

//Here divider is 10 low bits and multiplier is next 10 bits

#define PLL_MULDIV_MASK	    0xFFF00000	
#define PLL_MUL_MASK        0xFFF003FF 
#define PLL_DIV_MASK        0xFFFFFC00


/* Model Specific Registers. MSRs are 64-bit. For     */
/* writes (wrmsr - opcode 0f 30), the register number is placed in ecx, and   */
/* the value to write is placed in edx:eax. For reads (rdmsr - opcode 0f 32), */
/* the register number is placed in ecx, and the data is returned in edx:eax. */

#define MSR_FIDVID_CTL      0xc0010041
#define MSR_FIDVID_STATUS   0xc0010042

#define K8_MIN_VID	    0x1E
#define K8_MAX_VID          0
#define K8_MAX_FID	    0x2A
#define K8_MIN_FID	    0
#define DEF_STP_GNT_TO	    0x800			

/* Field definitions within the FID VID Low Control MSR : */

#define MSR_C_LO_INIT_FID_VID     0x00010000
#define MSR_C_LO_NEW_VID          0x00001F00
#define MSR_C_LO_NEW_FID          0x0000003F

/* Field definitions within the FID VID High Control MSR : */

#define MSR_C_HI_STP_GNT_TO 	  0x000FFFFF

/* Field definitions within the FID VID Low Status MSR : */

#define MSR_S_LO_CHANGE_PENDING   0x80000000   /* cleared when completed */
#define MSR_S_LO_MAX_RAMP_VID     0x1F000000
#define MSR_S_LO_MAX_FID          0x003F0000
#define MSR_S_LO_START_FID        0x00003F00
#define MSR_S_LO_CURRENT_FID      0x0000003F

/* Field definitions within the FID VID High Status MSR : */
#define MSR_S_HI_MIN_WORKING_VID  0x1F000000
#define MSR_S_HI_MAX_WORKING_VID  0x001F0000
#define MSR_S_HI_START_VID        0x00001F00
#define MSR_S_HI_CURRENT_VID      0x0000001F

/*----------------------------*/
/*Conversions*/
#define FID_to_MUL(FID)	        	(FID/2+4) 
#define VID_to_VOLTAGE(VID)		(1550-25*VID) 


#define GET_S_LO_CURRENT_FID(MSR_S_LO)     (MSR_S_LO&MSR_S_LO_CURRENT_FID);
#define GET_S_LO_START_FID(MSR_S_LO)       ((MSR_S_LO&MSR_S_LO_START_FID)>>8);
#define GET_S_LO_MAX_FID(MSR_S_LO)         ((MSR_S_LO&MSR_S_LO_MAX_FID)>>16);
#define GET_S_LO_MAX_RAMP_VID(MSR_S_LO)    ((MSR_S_LO&MSR_S_LO_MAX_RAMP_VID)>>24);
#define GET_S_LO_CHANGE_PENDING(MSR_S_LO)  ((MSR_S_LO&MSR_S_LO_CHANGE_PENDING)>>31);

#define GET_S_HI_CURRENT_VID(MSR_S_HI)     (MSR_S_HI&MSR_S_HI_CURRENT_VID);
#define GET_S_HI_START_VID(MSR_S_HI)       ((MSR_S_HI&MSR_S_HI_START_VID)>>8);
#define GET_S_HI_MAX_VID(MSR_S_HI)         ((MSR_S_HI&MSR_S_HI_MAX_WORKING_VID)>>16);
#define GET_S_HI_MIN_VID(MSR_S_HI)         ((MSR_S_HI&MSR_S_HI_MIN_WORKING_VID)>>24);


/*CK804 ISA Bridge*/
static struct pci_dev *CK804_ISA_BRIDGE_DEV;


static unsigned short PLL_PORT;
static unsigned short PLL_ENABLE_PORT;
static unsigned int   START_FSB;
static unsigned int START_VID=UNSPECIFIED;
static unsigned int START_MUL=UNSPECIFIED;
static cpumask_t    AFFECTED_CPUS;
//ALG -specified algorithm used for FSB changing
// 0-nTune type, 1-clockgen type, 2,3 -clockgen with divider 0x200,0x300
static unsigned int ALG=0;
/*Set control register*/
inline void SET_C_LO_NEW_FID(u32 *msr_c_lo, u32 fid)
{
	(*msr_c_lo)&=(~MSR_C_LO_NEW_FID);	
	(*msr_c_lo)|=fid;
}

inline void SET_C_LO_NEW_VID(u32 *msr_c_lo, u32 vid)
{
	(*msr_c_lo)&=(~MSR_C_LO_NEW_VID);	
	(*msr_c_lo)|=(vid<<8);
}

inline void SET_C_LO_INIT(u32 *msr_c_lo)
{
	(*msr_c_lo)|=(1<<16);		
}

inline void SET_C_HI_STP_GNT_TO(u32 *msr_c_hi, u32 stp_gnt_to)
{
	(*msr_c_hi)&=(~MSR_C_HI_STP_GNT_TO);	
	(*msr_c_hi)|=stp_gnt_to;
}

struct K8_STATUS_MSR
{
	int cpu;
	u32 current_fid;
	u32 start_fid;
	u32 max_fid;
	u32 max_ramp_vid;
	u32 change_pending;
	u32 current_vid;
	u32 start_vid;
	u32 max_vid;
	u32 min_vid;	
};
void PRINT_STATUS_MSR(struct K8_STATUS_MSR status_msr)
{
    printk(KERN_INFO "-------------------------------\n");
    printk(KERN_INFO "cpufreq: CPU %d:\n",status_msr.cpu);

    printk(KERN_INFO "cpufreq: current multiplier=%d\n", FID_to_MUL(status_msr.current_fid));
    printk(KERN_INFO "cpufreq: start multiplier=%d\n", FID_to_MUL(status_msr.start_fid));
    printk(KERN_INFO "cpufreq: max multiplier=%d\n", FID_to_MUL(status_msr.start_fid));
    
    printk(KERN_INFO "cpufreq: current voltage=%d mv \n", VID_to_VOLTAGE(status_msr.current_vid));
    printk(KERN_INFO "cpufreq: start voltage=%d mv \n", VID_to_VOLTAGE(status_msr.start_vid));
    printk(KERN_INFO "cpufreq: max voltage=%d mv \n", VID_to_VOLTAGE(status_msr.max_vid));
    printk(KERN_INFO "cpufreq: min voltage=%d mv \n", VID_to_VOLTAGE(status_msr.min_vid));
    printk(KERN_INFO "cpufreq: max ramp voltage=%d mv \n", VID_to_VOLTAGE(status_msr.max_ramp_vid));    
    printk(KERN_INFO "cpufreq: change pending=%d\n", status_msr.change_pending);
    
    printk(KERN_INFO "-------------------------------\n");
    
}
struct K8_CTL_MSR
{
	int cpu;
	u32 fid;
	u32 vid;
	u32 stp_gnt_to;
	u32 init_bit;
};
struct PLL
{
	u32 mul;
	u32 div;
	u32 pll;
};
inline void SET_MULDIV(struct PLL *pll)
{
	pll->mul=((pll->pll)>>10)&0x3FF;
	pll->div=(pll->pll)&0x3FF;
}
inline void SET_PLL(struct PLL *pll)
{
	(pll->pll)&=PLL_MULDIV_MASK;
	(pll->pll)|=((pll->mul)<<10);
	(pll->pll)|=(pll->div);
}
void ENABLE_PLL(void)
{
	u8 enable;
	enable=inb_p(PLL_ENABLE_PORT);
	/*enable: 0-enable, 1-disable. Set enable bit if PLL disabled*/
	if(enable&0x1)
	{
		outb_p(enable&(~0x1),PLL_ENABLE_PORT);
#ifdef PLL_DELAY
		mdelay(PLL_DELAY); 
#endif
	}
}

void READ_PLL(struct PLL *pll)
{
	pll->pll=inl_p(PLL_PORT);
	SET_MULDIV(pll);
}
void WRITE_PLL(struct PLL pll)
{
	outl_p(pll.pll, PLL_PORT); 
#ifdef PLL_DELAY
	mdelay(PLL_DELAY); 
#endif
}

inline int ABS(int val)
{
	if(val>=0)
		return val;
	else
		return -val;
}

u32 FIND_BEST_PLL(u32 target_fsb, struct PLL *pll)
{
	u32 best_fsb=UNSPECIFIED;
	u32 tmp;
	u32 imul,idiv;
	for(imul=PLL_MIN_MUL;imul<=0x3FF;imul++)
		for(idiv=PLL_MIN_DIV;idiv<=0xFF;idiv++)
		{
			tmp=PLL_REF_CLOCK*imul/idiv;
			if ( ABS(best_fsb-target_fsb)>ABS(tmp-target_fsb) )
			{
				best_fsb=tmp;
				pll->mul=imul;
				pll->div=idiv;
			}				
		}
		
	SET_PLL(pll);			
	
	return best_fsb;
}

inline u32 MUL_to_FID(u32 mul)
{
	u32 fid;
	if(mul<FID_to_MUL(K8_MIN_FID))
		fid=K8_MIN_FID;
	else
		if(mul>FID_to_MUL(K8_MAX_FID))
			fid=K8_MAX_FID;
		else
			fid=2*(mul-4);
	return	fid; 			
}

inline u32 VOLTAGE_to_VID(u32 voltage)
{
	u32 vid;
	if(voltage<VID_to_VOLTAGE(K8_MIN_VID))
		vid=K8_MIN_VID;
	else
		if(voltage>VID_to_VOLTAGE(K8_MAX_VID))
			vid=K8_MAX_VID;
		else
			vid=(1550-voltage)/25;
	return	vid; 			
}


/* Shamelessly ripped from "msr.c" kernel driver
*/
/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2000 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 675 Mass Ave, Cambridge MA 02139,
 *   USA; either version 2 of the License, or (at your option) any later
 *   version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

/* Note: "err" is handled in a funny way below.  Otherwise one version
   of gcc or another breaks. */
static inline int wrmsr_eio(u32 reg, u32 eax, u32 edx)
{
	int err;

	asm volatile ("1:	wrmsr\n"
		      "2:\n"
		      ".section .fixup,\"ax\"\n"
		      "3:	movl %4,%0\n"
		      "	jmp 2b\n"
		      ".previous\n"
		      ".section __ex_table,\"a\"\n"
		      "	.align 8\n" "	.quad 1b,3b\n" ".previous":"=&bDS" (err)
		      :"a"(eax), "d"(edx), "c"(reg), "i"(-EIO), "0"(0));

	return err;
}
static inline int rdmsr_eio(u32 reg, u32 *eax, u32 *edx)
{
	int err;

	asm volatile ("1:	rdmsr\n"
		      "2:\n"
		      ".section .fixup,\"ax\"\n"
		      "3:	movl %4,%0\n"
		      "	jmp 2b\n"
		      ".previous\n"
		      ".section __ex_table,\"a\"\n"
		      "	.align 8\n"
		      "	.quad 1b,3b\n"
		      ".previous":"=&bDS" (err), "=a"(*eax), "=d"(*edx)
		      :"c"(reg), "i"(-EIO), "0"(0));

	return err;
}
#ifdef CONFIG_SMP
struct msr_command {
	int cpu;
	int err;
	u32 reg;
	u32 data[2];
};
static void msr_smp_wrmsr(void *cmd_block)
{
	struct msr_command *cmd = (struct msr_command *)cmd_block;

	if (cmd->cpu == smp_processor_id())
		cmd->err = wrmsr_eio(cmd->reg, cmd->data[0], cmd->data[1]);
}
static void msr_smp_rdmsr(void *cmd_block)
{
	struct msr_command *cmd = (struct msr_command *)cmd_block;

	if (cmd->cpu == smp_processor_id())
		cmd->err = rdmsr_eio(cmd->reg, &cmd->data[0], &cmd->data[1]);
}
static int do_wrmsr(int cpu, u32 reg, u32 eax, u32 edx)
{
	struct msr_command cmd;
	int ret;

	preempt_disable();
	if (cpu == smp_processor_id()) {
		ret = wrmsr_eio(reg, eax, edx);
	} else {
		cmd.cpu = cpu;
		cmd.reg = reg;
		cmd.data[0] = eax;
		cmd.data[1] = edx;

		smp_call_function(msr_smp_wrmsr, &cmd, 1, 1);
		ret = cmd.err;
	}
	preempt_enable();
	return ret;
}
static  int do_rdmsr(int cpu, u32 reg, u32 * eax, u32 * edx)
{
	struct msr_command cmd;
	int ret;

	preempt_disable();
	if (cpu == smp_processor_id()) {
		ret = rdmsr_eio(reg, eax, edx);
	} else {
		cmd.cpu = cpu;
		cmd.reg = reg;

		smp_call_function(msr_smp_rdmsr, &cmd, 1, 1);

		*eax = cmd.data[0];
		*edx = cmd.data[1];

		ret = cmd.err;
	}
	preempt_enable();
	return ret;
}
#else				/* ! CONFIG_SMP */
static int do_wrmsr(int cpu, u32 reg, u32 eax, u32 edx)
{
	return wrmsr_eio(reg, eax, edx);
}
static int do_rdmsr(int cpu, u32 reg, u32 *eax, u32 *edx)
{
	return rdmsr_eio(reg, eax, edx);
}
#endif				/* ! CONFIG_SMP */
/*----------End of msr.c kernel driver-------------*/
